import bson.json_util
import dateutil.parser
from functools import wraps
import inspect
import jsonschema
import os
import cherrypy
from collections import OrderedDict
import logging

from girder import constants
from girder.api.rest import getCurrentUser, getBodyJson
from girder.constants import SortDir, VERSION
from girder.exceptions import RestException
from girder.models.setting import Setting
from girder.settings import SettingKey
from girder.utility import toBool
from girder.utility.model_importer import ModelImporter
from girder.utility.webroot import WebrootBase
from . import docs, access
from .rest import Resource, getApiUrl, getUrlParts

from inspect import signature, Parameter

SWAGGER_VERSION = '2.0'
logger = logging.getLogger(__name__)


def _walkTree(node, path=()):
    # This will infinitely recurse if anything mounted under the API root
    # has a cyclical reference. That's bad.
    routeMap = {}
    for k, v in vars(node).items():
        if isinstance(v, Resource):
            full_path = list(path)
            full_path.append(k)
            routeMap[v] = full_path

        if hasattr(v, 'exposed'):
            new_path = list(path)
            new_path.append(k)
            routeMap.update(_walkTree(v, new_path))

    return routeMap


class Description:
    """
    This class provides convenient chainable semantics to allow api route
    handlers to describe themselves to the documentation. A route handler
    function can apply the :py:class:`girder.api.describe.describeRoute`
    decorator to itself (called with an instance of this class) in order to
    describe itself.
    """

    # Data Type map from common name or type to (type, format)
    # See Data Type spec:
    #   https://github.com/OAI/OpenAPI-Specification/blob/
    #   0122c22e7fb93b571740dd3c6e141c65563a18be/versions/2.0.md#data-types
    _dataTypeMap = {
        # Primitives
        'integer': ('integer', 'int32'),
        'long': ('integer', 'int64'),
        'number': ('number', None),
        'float': ('number', 'float'),
        'double': ('number', 'double'),
        'string': ('string', None),
        'byte': ('string', 'byte'),
        'binary': ('string', 'binary'),
        'boolean': ('boolean', None),
        'date': ('string', 'date'),
        'dateTime': ('string', 'date-time'),
        'password': ('string', 'password'),
        'file': ('file', None)
    }

    def __init__(self, summary):
        self._summary = summary
        self._params = []
        self._responses = {}
        self._consumes = []
        self._produces = []
        self._responseClass = None
        self._responseClassArray = False
        self._notes = None
        self._deprecated = False
        self.hasPagingParams = False
        self.modelParams = {}
        self.jsonParams = {}

    def asDict(self):
        """
        Returns this description object as an appropriately formatted dict
        """
        # Responses Object spec:
        # The Responses Object MUST contain at least one response code, and it
        # SHOULD be the response for a successful operation call.
        if '200' not in self._responses:
            self._responses['200'] = {
                'description': 'Success'
            }
        if self._responseClass is not None:
            schema = {
                '$ref': '#/definitions/%s' % self._responseClass
            }
            if self._responseClassArray:
                schema = {
                    'type': 'array',
                    'items': schema
                }
            self._responses['200']['schema'] = schema

        resp = {
            'summary': self._summary,
            'responses': self._responses
        }

        if self._params:
            resp['parameters'] = self._params

        if self._notes is not None:
            resp['description'] = self._notes

        if self._consumes:
            resp['consumes'] = self._consumes

        if self._produces:
            # swagger has a bug where not all appropriate mime types are
            # considered to be binary (see
            # https://github.com/swagger-api/swagger-ui/issues/1605).  If we
            # have specified zip format, replace it with
            # application/octet-stream
            #   Reduce the list of produces values to unique values,
            # maintaining the order.
            produces = list(OrderedDict.fromkeys([
                'application/octet-stream' if item in ('application/zip', )
                else item for item in self._produces]))
            resp['produces'] = produces

        if self._deprecated:
            resp['deprecated'] = True

        return resp

    def responseClass(self, obj, array=False):
        self._responseClass = obj
        self._responseClassArray = array
        return self

    def _validateParamInfo(self, dataType, paramType, name):
        """
        Helper to convert and validate the dataType and paramType.
        Prints warnings if invalid values were passed.
        """
        # Legacy data type conversions
        if dataType == 'int':
            dataType = 'integer'

        # Parameter Object spec:
        # If type is "file", then the swagger "consumes" field MUST be either
        # "multipart/form-data", "application/x-www-form-urlencoded" or both
        # and the parameter MUST be in "formData".
        if dataType == 'file':
            paramType = 'formData'

        # Get type and format from common name
        dataTypeFormat = None
        if dataType in self._dataTypeMap:
            dataType, dataTypeFormat = self._dataTypeMap[dataType]
        # If we are dealing with the body then the dataType might be defined
        # by a schema added using addModel(...), we don't know for sure as we
        # don't know the resource name here to look it up.
        elif paramType != 'body':
            logger.warning(
                'WARNING: Invalid dataType "%s" specified for parameter names "%s"' %
                (dataType, name))

        # Parameter Object spec:
        # Since the parameter is not located at the request body, it is limited
        # to simple types (that is, not an object).
        if paramType != 'body' and dataType not in (
                'string', 'number', 'integer', 'long', 'boolean', 'array', 'file', 'float',
                'double', 'date', 'dateTime'):
            logger.warning(
                'WARNING: Invalid dataType "%s" specified for parameter "%s"' % (dataType, name))

        if paramType == 'form':
            paramType = 'formData'

        return dataType, dataTypeFormat, paramType

    def param(self, name, description, paramType='query', dataType='string',
              required=True, enum=None, default=None, strip=False, lower=False, upper=False):
        """
        This helper will build a parameter declaration for you. It has the most
        common options as defaults, so you won't have to repeat yourself as much
        when declaring the APIs.

        Note that we could expose more parameters from the Parameter Object
        spec, for example: format, allowEmptyValue, minimum, maximum, pattern,
        uniqueItems.

        :param name: name of the parameter used in the REST query.
        :param description: explanation of the parameter.
        :param paramType: how is the parameter sent.  One of 'query', 'path',
                          'body', 'header', or 'formData'.
        :param dataType: the data type expected in the parameter. This is one
                         of 'integer', 'long', 'float', 'double', 'string',
                         'byte', 'binary', 'boolean', 'date', 'dateTime',
                         'password', or 'file'.
        :param required: True if the request will fail if this parameter is not
                         present, False if the parameter is optional.
        :param enum: a fixed list of possible values for the field.
        :type enum: `list`
        :param strip: For string types, set this to True if the string should be
            stripped of white space.
        :type strip: bool
        :param lower: For string types, set this to True if the string should be
            converted to lowercase.
        :type lower: bool
        :param upper: For string types, set this to True if the string should be
            converted to uppercase.
        :type upper: bool
        """
        dataType, format, paramType = self._validateParamInfo(dataType, paramType, name)

        param = {
            'name': name,
            'description': description,
            'in': paramType,
            'required': required
        }

        if dataType == 'string':
            param['_strip'] = strip
            param['_lower'] = lower
            param['_upper'] = upper

        if paramType == 'body':
            if format is None:
                param['schema'] = {
                    '$ref': '#/definitions/%s' % dataType
                }
            else:
                param['schema'] = {
                    'type': dataType,
                    'format': format,
                }
        else:
            param['type'] = dataType
            if format is not None:
                param['format'] = format

        if enum:
            param['enum'] = enum

        if default is not None:
            param['default'] = default

        self._params.append(param)
        return self

    def modelParam(self, name, description=None, model=None, destName=None, paramType='path',
                   plugin='_core', level=None, required=True, force=False, exc=True,
                   requiredFlags=None, **kwargs):
        """
        This should be used in lieu of ``param`` if the parameter is a model ID
        and the model should be loaded and passed into the route handler. For example,
        if you have a route like ``GET /item/:id``, you could do:

        >>> from girder.models.item import Item
        >>> modelParam('id', model=Item, level=AccessType.READ)

        Which would cause the ``id`` parameter in the path to be mapped to an
        item model parameter named ``item``, and ensure that the calling user
        has at least ``READ`` access on that item. For parameters passed in
        the query string or form data, for example a request like
        ``POST /item?folderId=...``, you must specify the ``paramType``.

        >>> modelParam('folderId', 'The ID of the parent folder.', model=Folder,
        ...            level=AccessType.WRITE, paramType='query')

        Note that in the above example, ``model`` is omitted; in this case, the
        model is inferred to be ``'folder'`` from the parameter name ``'folderId'``.

        :param name: The name passed in via the request, e.g. 'id'.
        :type name: str
        :param description: The description of the parameter. If not passed, defaults
            to "The ID of the <model>."
        :type description: str
        :param destName: The kwarg name after model loading, e.g. 'folder'. Defaults
            to the value of the model parameter.
        :type destName: str
        :param paramType: how is the parameter sent.  One of 'query', 'path',
            'body', 'header', or 'formData'.
        :param model: The model class to use for loading, or a name, e.g. 'folder'. If not passed,
            defaults to stripping the last two characters from the name, such that e.g. 'folderId'
            would make the model become 'folder'.
        :type model: class or str
        :param plugin: Plugin name, if loading a plugin model. Only used when the ``model``
            param is a string rather than a class.
        :type plugin: str
        :param level: Access level, if this is an access controlled model.
        :type level: AccessType
        :param required: Whether this parameter is required.
        :type required: bool
        :param force: Force loading of the model (skip access check).
        :type force: bool
        :param exc: Whether an exception should be raised for a nonexistent resource.
        :type exc: bool
        :param requiredFlags: Access flags that are required on the object being loaded.
        :type requiredFlags: str or list/set/tuple of str or None
        """
        if model is None:
            model = name[:-2]  # strip off "Id"

        isModelClass = inspect.isclass(model)

        if description is None:
            description = 'The ID of the document.'

        self.param(name=name, description=description, paramType=paramType, required=required)

        self.modelParams[name] = {
            'destName': destName,
            'level': level,
            'force': force,
            'model': model,
            'plugin': plugin,
            'isModelClass': isModelClass,
            'exc': exc,
            'required': required,
            'requiredFlags': requiredFlags,
            'kwargs': kwargs
        }

        return self

    def jsonParam(self, name, description, paramType='query', dataType='string', required=True,
                  default=None, requireObject=False, requireArray=False, schema=None):
        """
        Specifies a parameter that should be processed as JSON.

        :param requireObject: Whether the value must be a JSON object / Python dict.
        :type requireObject: bool
        :param requireArray: Whether the value must be a JSON array / Python list.
        :type requireArray: bool
        :param schema: A JSON schema that will be used to validate the parameter value. If
            this is passed, it overrides any ``requireObject`` or ``requireArray`` values
            that were passed.
        :type schema: dict
        """
        if default is not None and default != ():
            default = bson.json_util.dumps(default)

        self.param(
            name=name, description=description, paramType=paramType, dataType=dataType,
            required=required, default=default)

        self.jsonParams[name] = {
            'requireObject': requireObject,
            'requireArray': requireArray,
            'schema': schema
        }

        return self

    def pagingParams(self, defaultSort, defaultSortDir=SortDir.ASCENDING, defaultLimit=50):
        """
        Adds the limit, offset, sort, and sortdir parameter documentation to
        this route handler.

        :param defaultSort: The default field used to sort the result set.
        :type defaultSort: str
        :param defaultSortDir: Sort order: -1 or 1 (desc or asc)
        :type defaultSortDir: int
        :param defaultLimit: The default page size.
        :type defaultLimit: int
        """
        self.param(
            'limit', 'Result set size limit.', default=defaultLimit, required=False, dataType='int')
        self.param('offset', 'Offset into result set.', default=0, required=False, dataType='int')

        if defaultSort is not None:
            self.param(
                'sort', 'Field to sort the result set by.', default=defaultSort, required=False,
                strip=True)
            self.param(
                'sortdir', 'Sort order: 1 for ascending, -1 for descending.',
                required=False, dataType='integer', enum=[SortDir.ASCENDING, SortDir.DESCENDING],
                default=defaultSortDir)

        self.hasPagingParams = True
        return self

    def consumes(self, value):
        self._consumes.append(value)
        return self

    def produces(self, value):
        if isinstance(value, (list, tuple)):
            self._produces.extend(value)
        else:
            self._produces.append(value)
        return self

    def notes(self, notes):
        self._notes = notes
        return self

    def errorResponse(self, reason='A parameter was invalid.', code=400):
        """
        This helper will build an errorResponse declaration for you. Many
        endpoints will be able to use the default parameter values for one of
        their responses.

        :param reason: The reason or list of reasons why the error occurred.
        :type reason: `str, list, or tuple`
        :param code: HTTP status code.
        :type code: int
        """
        code = str(code)

        # Combine list of reasons into a single string.
        # swagger-ui renders the description using Markdown.
        if not isinstance(reason, str):
            reason = '\n\n'.join(reason)

        if code in self._responses:
            self._responses[code]['description'] += '\n\n' + reason
        else:
            self._responses[code] = {
                'description': reason
            }

        return self

    def deprecated(self):
        """
        Mark the route as deprecated.
        """
        self._deprecated = True
        return self

    @property
    def params(self):
        return self._params


class ApiDocs(WebrootBase):
    """
    This serves up the Swagger page.
    """

    def __init__(self, templatePath=None):
        if not templatePath:
            templatePath = os.path.join(constants.PACKAGE_DIR, 'api', 'api_docs.mako')
        super().__init__(templatePath)

    def _renderHTML(self):
        self.vars['brandName'] = Setting().get(SettingKey.BRAND_NAME)
        urlRoot = (os.environ.get('GIRDER_URL_ROOT') or '').rstrip('/')
        if not urlRoot.startswith('.') and ':' not in urlRoot and urlRoot:
            urlRoot = '/' + urlRoot
        self.vars['urlRoot'] = urlRoot
        return super()._renderHTML()


class Describe(Resource):
    def __init__(self):
        super().__init__()
        self.route('GET', (), self.listResources, nodoc=True)

    def _addSwaggerValues(self, paths, definitions):
        for path in paths.values():
            for method in path.values():
                for param in method.get('parameters', []):
                    if param.get('in', '') == 'formData' and 'consumes' not in method:
                        method['consumes'] = [
                            'multipart/form-data', 'application/x-www-form-urlencoded']
                    if param.get('default') == ():
                        param['default'] = bson.json_util.dumps(param['default'])
                for param in (method.get('parameters', []) + list(
                        method.get('responses', {}).values())):
                    if 'schema' not in param:
                        continue
                    schema = param['schema']
                    schema = schema.get('items', schema)
                    defin = schema.get('$ref', '').split('#/definitions/')
                    if (len(defin) == 2 and defin[1] not in definitions
                            and defin[1] not in {'string', 'boolean', 'integer', 'number'}):
                        definitions[defin[1]] = {'type': 'object'}

    @access.public
    def listResources(self, params):
        # Paths Object
        paths = {}

        # Definitions Object
        definitions = dict(**docs.models[None])

        wsgiRoot = cherrypy.request.app.root

        # This is due to the idiosyncrasy of mounting the API root at both {core_girder_route}/api
        # as well as under /api. Depending which one is being used, the API root object could be
        # either at .api.v1 or just at .v1 due to the way it's mounted in server.py.
        try:
            apiRoot = wsgiRoot.api.v1
        except AttributeError:
            apiRoot = wsgiRoot.v1

        routeMap = _walkTree(apiRoot)

        # List of Tag Objects
        tags = []

        for resource in sorted(docs.routes.keys(), key=str):
            # Update Definitions Object
            if resource in docs.models:
                for name, model in docs.models[resource].items():
                    definitions[name] = model

            prefixPath = None
            tag = resource
            if isinstance(resource, Resource):
                if resource not in routeMap:
                    raise RestException('Resource not mounted: %s' % resource)
                prefixPath = routeMap[resource]
                tag = prefixPath[0]

            # Tag Object
            tags.append({
                'name': tag
            })

            for route, methods in docs.routes[resource].items():
                # Path Item Object
                pathItem = {}
                for method, operation in methods.items():
                    operation = operation.copy()
                    if 'parameters' in operation:
                        operation['parameters'] = [
                            {k: v for k, v in param.items() if not k.startswith('_')}
                            for param in operation['parameters']]
                    # Operation Object
                    pathItem[method.lower()] = operation
                    if prefixPath:
                        operation['tags'] = prefixPath[:1]

                if prefixPath:
                    route = '/'.join([''] + prefixPath + [route[1:]])

                paths[route] = pathItem

        self._addSwaggerValues(paths, definitions)

        apiUrl = getApiUrl(preferReferer=True)
        urlParts = getUrlParts(apiUrl)
        host = urlParts.netloc
        basePath = urlParts.path
        brandName = Setting().get(SettingKey.BRAND_NAME) or 'Girder'

        paths['/user/authentication']['get']['security'] = [{'basicAuth': []}]

        return {
            'swagger': SWAGGER_VERSION,
            'info': {
                'title': f'{brandName} REST API',
                'version': VERSION['release'],
                'description':
                    'Below you will find the list of all of the resource '
                    f'types exposed by the {brandName} RESTful Web API. '
                    'Click any of the resource links to open up a list of '
                    'all available endpoints related to each resource type. '
                    '\n\nFor endpoints that require authentication, a Girder '
                    'token is required. This can be obtained via POST '
                    '`/api_key/token` using an API key, or from GET '
                    '`/user/authentication`. The latter endpoint can be '
                    'authenticated via Basic Auth, if enabled in the system.',
            },
            'host': host,
            'basePath': basePath,
            'tags': tags,
            'paths': paths,
            'definitions': definitions,
            'securityDefinitions': {
                'Girder-Token': {
                    'type': 'apiKey',
                    'in': 'header',
                    'name': 'Girder-Token',
                },
                'basicAuth': {
                    'type': 'basic',
                },
            },
            'security': [{'Girder-Token': []}],
        }


class describeRoute:  # noqa: class name
    def __init__(self, description):
        """
        This returns a decorator to set the API documentation on a route
        handler. Pass the Description object (or None) that you want to use to
        describe this route. It should be used like the following example:

        .. code-block:: python

            @describeRoute(
                Description('Do something')
               .param('foo', 'Some parameter', ...)
            )
            def routeHandler(...)

        :param description: The description for the route.
        :type description: :py:class:`girder.api.describe.Description` or None
        """
        self.description = description

    def __call__(self, fun):
        fun.description = self.description
        return fun


class autoDescribeRoute(describeRoute):  # noqa: class name
    def __init__(self, description, hide=False):
        """
        Like describeRoute, but this decorator also controls behavior of the
        underlying method. It handles parameter validation and transformation
        based on the Description object passed.

        :param description: The description object.
        :type description: Description
        :param hide: Set to True if this route should not appear in the swagger listing.
        :type hide: bool
        """
        super().__init__(description=description)
        self.hide = hide

    def _passArg(self, fun, kwargs, name, val):
        """
        This helper passes the arguments to the underlying function if the function
        has an argument with the given name. Otherwise, it adds it into the "params"
        argument, which is a dictionary containing other parameters.

        :param fun: The wrapped route handler function
        :type fun: callable
        :param name: The name of the argument to set
        :type name: str
        :param kwargs: The arguments to be passed down to the function.
        :type kwargs: dict
        :param val: The value of the argument to set
        """
        if name in self._funNamedArgs or self._funHasKwargs:
            kwargs[name] = val
            kwargs['params'].pop(name, None)
        else:
            kwargs['params'][name] = val

    def _mungeKwargs(self, kwargs, fun):
        """
        Performs final modifications to the kwargs passed into the wrapped function.
        Combines the sort/sortdir params appropriately for consumption by the model
        layer, and only passes the "params" catch-all dict if there is a corresponding
        kwarg for it in the wrapped function.
        """
        if self.description.hasPagingParams and 'sort' in kwargs:
            sortdir = kwargs.pop('sortdir', None) or kwargs['params'].pop('sortdir', None)
            kwargs['sort'] = [(kwargs['sort'], sortdir)]

        if 'params' not in self._funNamedArgs and not self._funHasKwargs:
            kwargs.pop('params', None)

    def _inspectFunSignature(self, fun):
        self._funNamedArgs = set()
        self._funHasKwargs = False
        for funParam in signature(fun).parameters.values():
            if funParam.kind in {Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
                # POSITIONAL_OR_KEYWORD are basic positional parameters
                # KEYWORD_ONLY are named parameters that appear after a * in Python 3
                self._funNamedArgs.add(funParam.name)
            elif funParam.kind == Parameter.VAR_KEYWORD:
                # VAR_KEYWORD is the **kwargs parameter
                self._funHasKwargs = True

    @staticmethod
    def _destName(info, model):
        destName = info['destName']
        if destName is None:
            if info['isModelClass']:
                destName = model.name
            else:
                destName = info['model']

        return destName

    def __call__(self, fun):
        self._inspectFunSignature(fun)

        @wraps(fun)
        def wrapped(*args, **kwargs):
            """
            Transform any passed params according to the spec, or
            fill in default values for any params not passed.
            """
            # Combine path params with form/query params into a single lookup table
            params = {k: v for k, v in kwargs.items() if k != 'params'}
            params.update(kwargs.get('params', {}))

            kwargs['params'] = kwargs.get('params', {})

            for descParam in self.description.params:
                # We need either a type or a schema ( for message body )
                if 'type' not in descParam and 'schema' not in descParam:
                    continue

                name = descParam['name']
                model = self._getModel(name, self.description.modelParams)
                if name in params:
                    if name in self.description.jsonParams:
                        info = self.description.jsonParams[name]
                        val = self._loadJson(name, info, params[name])
                        self._passArg(fun, kwargs, name, val)
                    elif name in self.description.modelParams:
                        info = self.description.modelParams[name]
                        kwargs.pop(name, None)  # Remove from path params
                        val = self._loadModel(name, info, params[name], model)
                        self._passArg(fun, kwargs, self._destName(info, model), val)
                    else:
                        val = self._validateParam(name, descParam, params[name])
                        self._passArg(fun, kwargs, name, val)
                elif descParam['in'] == 'body':
                    if name in self.description.jsonParams:
                        info = self.description.jsonParams[name].copy()
                        info['required'] = descParam['required']
                        val = self._loadJsonBody(name, info)
                        self._passArg(fun, kwargs, name, val)
                    else:
                        self._passArg(fun, kwargs, name, cherrypy.request.body)
                elif descParam['in'] == 'header':
                    continue  # For now, do nothing with header params
                elif 'default' in descParam:
                    self._passArg(fun, kwargs, name, descParam['default'])
                elif descParam['required']:
                    raise RestException('Parameter "%s" is required.' % name)
                else:
                    # If required=False but no default is specified, use None
                    if name in self.description.modelParams:
                        info = self.description.modelParams[name]
                        kwargs.pop(name, None)  # Remove from path params
                        self._passArg(fun, kwargs, info['destName'] or model.name, None)
                    else:
                        self._passArg(fun, kwargs, name, None)

            self._mungeKwargs(kwargs, fun)

            return fun(*args, **kwargs)

        if self.hide:
            wrapped.description = None
        else:
            wrapped.description = self.description
        return wrapped

    def _validateJsonType(self, name, info, val):
        if info.get('schema') is not None:
            try:
                jsonschema.validate(val, info['schema'])
            except jsonschema.ValidationError as e:
                raise RestException('Invalid JSON object for parameter %s: %s' % (
                    name, str(e)))
        elif info['requireObject'] and not isinstance(val, dict):
            raise RestException('Parameter %s must be a JSON object.' % name)
        elif info['requireArray'] and not isinstance(val, list):
            raise RestException('Parameter %s must be a JSON array.' % name)

    def _loadJsonBody(self, name, info):
        val = None
        if cherrypy.request.body.length == 0 and info['required']:
            raise RestException('JSON parameter %s must be passed in request body.' % name)
        elif cherrypy.request.body.length > 0:
            val = getBodyJson()
            self._validateJsonType(name, info, val)

        return val

    def _loadJson(self, name, info, value):
        try:
            val = bson.json_util.loads(value)
        except ValueError:
            raise RestException('Parameter %s must be valid JSON.' % name)

        self._validateJsonType(name, info, val)

        return val

    def _getModel(self, name, modelParams):
        if name not in self.description.modelParams:
            return
        info = self.description.modelParams[name]

        if info['isModelClass']:
            return info['model']()
        else:
            return ModelImporter.model(info['model'], info['plugin'])

    def _loadModel(self, name, info, id, model):
        if info['force']:
            doc = model.load(id, force=True, **info['kwargs'])
        elif info['level'] is not None:
            doc = model.load(id=id, level=info['level'], user=getCurrentUser(), **info['kwargs'])
        else:
            doc = model.load(id, **info['kwargs'])

        if doc is None and info['exc']:
            raise RestException('Invalid %s id (%s).' % (model.name, str(id)))

        if info['requiredFlags']:
            model.requireAccessFlags(doc, user=getCurrentUser(), flags=info['requiredFlags'])

        return doc

    def _handleString(self, name, descParam, value):
        if descParam['_strip']:
            value = value.strip()
        if descParam['_lower']:
            value = value.lower()
        if descParam['_upper']:
            value = value.upper()

        format = descParam.get('format')
        if format in ('date', 'date-time'):
            try:
                value = dateutil.parser.parse(value)
            except ValueError:
                raise RestException('Invalid date format for parameter %s: %s.' % (name, value))

            if format == 'date':
                value = value.date()

        return value

    def _handleInt(self, name, descParam, value):
        try:
            return int(value)
        except ValueError:
            raise RestException('Invalid value for integer parameter %s: %s.' % (name, value))

    def _handleNumber(self, name, descParam, value):
        try:
            return float(value)
        except ValueError:
            raise RestException('Invalid value for numeric parameter %s: %s.' % (name, value))

    def _validateParam(self, name, descParam, value):
        """
        Validates and transforms a single parameter that was passed. Raises
        RestException if the passed value is invalid.

        :param name: The name of the param.
        :type name: str
        :param descParam: The formal parameter in the Description.
        :type descParam: dict
        :param value: The value passed in for this param for the current request.
        :returns: The value transformed
        """
        type = descParam.get('type')

        # Coerce to the correct data type
        if type == 'string':
            value = self._handleString(name, descParam, value)
        elif type == 'boolean':
            value = toBool(value)
        elif type == 'integer':
            value = self._handleInt(name, descParam, value)
        elif type == 'number':
            value = self._handleNumber(name, descParam, value)

        # Enum validation (should be after type coercion)
        if 'enum' in descParam and value not in descParam['enum']:
            raise RestException('Invalid value for %s: "%s". Allowed values: %s.' % (
                name, value, ', '.join(str(v) for v in descParam['enum'])))

        return value



import time

from girder.exceptions import ValidationException, RestException
from girder.notification import Notification, ProgressState


class ProgressContext:
    """
    This class is a context manager that can be used to update progress in a way
    that rate-limits writes to the database and guarantees a flush when the
    context is exited. This is a no-op if "on" is set to False, which is
    meant as a convenience for callers. Any additional kwargs passed to this
    constructor are passed through to the ``initProgress`` method of the
    notification model.

    :param on: Whether to record progress.
    :type on: bool
    :param interval: Minimum time interval at which to write updates to the
        database, in seconds.
    :type interval: int or float
    :param user: The user creating this progress.
    :type user: dict
    :param title: The title for the task being tracked.
    :type title: str
    """

    def __init__(self, on, interval=0.5, **kwargs):
        self.on = on
        self.interval = interval
        self._lastFlush = time.time()

        if on:
            self.progress = Notification.initProgress(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, excType, excValue, traceback):
        if not self.on:
            return

        if excType is None and excValue is None:
            state = ProgressState.SUCCESS
            message = 'Done'
        else:
            state = ProgressState.ERROR
            message = 'Error'
            if isinstance(excValue, (ValidationException, RestException)):
                message = 'Error: ' + str(excValue)

        self.progress.updateProgress(state=state, message=message)

    def update(self, force: bool = False, increment: int = None, **kwargs):
        """
        Update the underlying progress record. This will only actually save
        to the database if at least self.interval seconds have passed since
        the last time the record was written to the database. Accepts the
        same kwargs as Notification.updateProgress.

        :param force: Whether we should force flushing the message, even if the minimum interval
            has not passed.
        :param increment: The amount to increment the current progress by.
        """
        if not self.on:
            return

        if (time.time() - self._lastFlush > self.interval) or force:
            self._lastFlush = time.time()
            self.progress.updateProgress(increment=increment, **kwargs)


noProgress = ProgressContext(on=False)



import os
import tempfile

import click
import uvicorn

from girder.constants import ServerMode
from girder.models.assetstore import Assetstore
from girder.models.file import File
from girder.utility import config

_default_db_url = os.environ.get('GIRDER_MONGO_URI', 'mongodb://localhost:27017/girder')
_default_mode = os.environ.get('GIRDER_SERVER_MODE', ServerMode.DEVELOPMENT)


@click.command(name='serve', short_help='Run the Girder server.', help='Run the Girder server.')
@click.option('--mode', type=click.Choice([
    ServerMode.PRODUCTION,
    ServerMode.DEVELOPMENT,
    ServerMode.TESTING
    ]), default=_default_mode, show_default=True, help='Specify the server mode')
@click.option('-d', '--database', default=_default_db_url,
              show_default=True, help='The database URI to connect to')
@click.option('-H', '--host', default='127.0.0.1',
              show_default=True, help='The interface to bind to')
@click.option('-p', '--port', type=int, default=8080,
              show_default=True, help='The port to bind to')
@click.option('--with-temp-assetstore', default=False, is_flag=True,
              help='Create a temporary assetstore for this server instance')
def main(mode: str, database: str, host: str, port: int, with_temp_assetstore: bool):
    # Must set these in env when using `reload=True`, due to uvicorn's use of subprocesses
    os.environ['GIRDER_SERVER_MODE'] = mode
    os.environ['GIRDER_MONGO_URI'] = database
    config.getConfig()['server'] = {'mode': mode}
    config.getConfig()['database']['uri'] = database

    def _run_app():
        uvicorn.run(
            'girder.asgi:app',
            host=host,
            port=port,
            reload=mode != ServerMode.PRODUCTION,
            server_header=False,
            date_header=False,
            log_level='warning',
        )

    if with_temp_assetstore:
        with tempfile.TemporaryDirectory() as tempdir:
            assetstore = Assetstore().createFilesystemAssetstore(
                name=tempdir,
                root=tempdir,
            )
            try:
                _run_app()
            finally:
                # Delete all files in the assetstore
                File().removeWithQuery({
                    'assetstoreId': assetstore['_id']
                })
                Assetstore().remove(assetstore)
    else:
        _run_app()



from ..describe import Description, autoDescribeRoute
from ..rest import Resource as BaseResource, setResponseHeader, setContentDisposition
from girder.constants import AccessType, TokenScope
from girder.exceptions import RestException
from girder.api import access
from girder.utility import parseTimestamp
from girder.utility.search import getSearchModeHandler
from girder.utility import ziputil
from girder.utility import path as path_util
from girder.utility.model_importer import ModelImporter
from girder.utility.progress import ProgressContext

# Plugins can modify this set to allow other types to be searched
allowedSearchTypes = {'collection', 'file', 'folder', 'group', 'item', 'user'}
allowedDeleteTypes = {'collection', 'file', 'folder', 'group', 'item', 'user'}


class Resource(BaseResource):
    """
    API Endpoints that deal with operations across multiple resource types.
    """

    def __init__(self):
        super().__init__()
        self.resourceName = 'resource'
        self.route('GET', ('search',), self.search)
        self.route('GET', ('lookup',), self.lookup)
        self.route('GET', (':id',), self.getResource)
        self.route('GET', (':id', 'path'), self.path)
        self.route('PUT', (':id', 'timestamp'), self.setTimestamp)
        self.route('GET', ('download',), self.download)
        self.route('POST', ('download',), self.download)
        self.route('PUT', ('move',), self.moveResources)
        self.route('POST', ('copy',), self.copyResources)
        self.route('DELETE', (), self.delete)

    @access.public
    @autoDescribeRoute(
        Description('Search for resources in the system.')
        .param('q', 'The search query.')
        .param('mode', 'The search mode. Can always use either a text search or a '
               'prefix-based search.', required=False, default='text')
        .jsonParam('types', 'A JSON list of resource types to search for, e.g. '
                   '["user", "folder", "item"].', requireArray=True)
        .param('level', 'Minimum required access level.', required=False,
               dataType='integer', default=AccessType.READ)
        .pagingParams(defaultSort=None, defaultLimit=10)
        .errorResponse('Invalid type list format.')
    )
    def search(self, q, mode, types, level, limit, offset):
        """
        Perform a search using one of the registered search modes.
        """
        level = AccessType.validate(level)
        user = self.getCurrentUser()
        handler = getSearchModeHandler(mode)
        if handler is None:
            raise RestException('Search mode handler %r not found.' % mode)
        results = handler(
            query=q,
            types=types,
            user=user,
            limit=limit,
            offset=offset,
            level=level
        )
        return results

    def _validateResourceSet(self, resources, allowedModels=None):
        """
        Validate a set of resources against a set of allowed models.
        Also ensures the requested resource set is not empty.
        # TODO jsonschema could replace this probably

        :param resources: The set of resources requested.
        :param allowedModels: if present, an iterable of models that may be
            included in the resources.
        """
        if allowedModels:
            invalid = set(resources.keys()) - set(allowedModels)
            if invalid:
                raise RestException('Invalid resource types requested: ' + ', '.join(invalid))
        count = sum([len(v) for v in resources.values()])
        if not count:
            raise RestException('No resources specified.')

    def _getResourceModel(self, kind, funcName=None):
        """
        Load and return a model with a specific function or throw an exception.

        :param kind: the name of the model to load
        :param funcName: a function name to ensure that each model contains.
        :returns: the loaded model.
        """
        try:
            model = ModelImporter.model(kind)
        except Exception:
            model = None
        if not model or (funcName and not hasattr(model, funcName)):
            raise RestException('Invalid resources format.')
        return model

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Look up a resource in the data hierarchy by path.')
        .param('path',
               'The path of the resource.  The path must be an absolute Unix '
               'path starting with either "/user/[user name]", for a user\'s '
               'resources or "/collection/[collection name]", for resources '
               'under a collection.')
        .errorResponse('Path is invalid.')
        .errorResponse('Path refers to a resource that does not exist.')
        .errorResponse('Read access was denied for the resource.', 403)
    )
    def lookup(self, path):
        return path_util.lookUpPath(path, self.getCurrentUser())['document']

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get path of a resource.')
        .param('id', 'The ID of the resource.', paramType='path')
        .param('type', 'The type of the resource (item, file, etc.).')
        .errorResponse('ID was invalid.')
        .errorResponse('Invalid resource type.')
        .errorResponse('Read access was denied for the resource.', 403)
    )
    def path(self, id, type):
        user = self.getCurrentUser()
        doc = self._getResource(id, type)
        if doc is None:
            raise RestException('Invalid resource id.')
        return path_util.getResourcePath(type, doc, user=user)

    @access.public(scope=TokenScope.DATA_READ, cookie=True)
    @autoDescribeRoute(
        Description('Download a set of items, folders, collections, and users '
                    'as a zip archive.')
        .notes('This route is also exposed via the POST method because the '
               'request parameters can be quite long, and encoding them in the '
               'URL (as is standard when using the GET method) can cause the '
               'URL to become too long, which causes errors.')
        .jsonParam('resources', 'A JSON-encoded set of resources to download. Each type is '
                   'a list of ids. For example: {"item": [(item id 1), (item id 2)], '
                   '"folder": [(folder id 1)]}.', requireObject=True)
        .param('includeMetadata', 'Include any metadata in JSON files in the '
               'archive.', required=False, dataType='boolean', default=False)
        .produces('application/zip')
        .errorResponse('Unsupported or unknown resource type.')
        .errorResponse('Invalid resources format.')
        .errorResponse('No resources specified.')
        .errorResponse('Resource not found.')
        .errorResponse('Read access was denied for a resource.', 403)
    )
    def download(self, resources, includeMetadata):
        """
        Returns a generator function that will be used to stream out a zip
        file containing the listed resource's contents, filtered by
        permissions.
        """
        user = self.getCurrentUser()
        self._validateResourceSet(resources)
        # Check that all the resources are valid, so we don't download the zip
        # file if it would throw an error.
        for kind in resources:
            model = self._getResourceModel(kind, 'fileList')
            for id in resources[kind]:
                if not model.load(id=id, user=user, level=AccessType.READ):
                    raise RestException('Resource %s %s not found.' % (kind, id))
        setResponseHeader('Content-Type', 'application/zip')
        setContentDisposition('Resources.zip')

        def stream():
            zip = ziputil.ZipGenerator()
            for kind in resources:
                model = ModelImporter.model(kind)
                for id in resources[kind]:
                    doc = model.load(id=id, user=user, level=AccessType.READ)
                    for (path, file) in model.fileList(
                            doc=doc, user=user, includeMetadata=includeMetadata, subpath=True):
                        yield from zip.addFile(file, path)
            yield zip.footer()
        return stream

    @access.user(scope=TokenScope.DATA_OWN)
    @autoDescribeRoute(
        Description('Delete a set of items, folders, or other resources.')
        .jsonParam('resources', 'A JSON-encoded set of resources to delete. Each '
                   'type is a list of ids.  For example: {"item": [(item id 1), '
                   '(item id2)], "folder": [(folder id 1)]}.', requireObject=True)
        .param('progress', 'Whether to record progress on this task.',
               default=False, required=False, dataType='boolean')
        .errorResponse('Unsupported or unknown resource type.')
        .errorResponse('Invalid resources format.')
        .errorResponse('No resources specified.')
        .errorResponse('Resource not found.')
        .errorResponse('Admin access was denied for a resource.', 403)
    )
    def delete(self, resources, progress):
        user = self.getCurrentUser()
        self._validateResourceSet(resources, allowedDeleteTypes)
        total = sum([len(resources[key]) for key in resources])
        with ProgressContext(
                progress, user=user, title='Deleting resources',
                message='Calculating size...') as ctx:
            ctx.update(total=total)
            current = 0
            for kind in resources:
                model = self._getResourceModel(kind, 'remove')
                for id in resources[kind]:
                    doc = model.load(id=id, user=user, level=AccessType.ADMIN, exc=True)

                    # Don't do a subtree count if we weren't asked for progress
                    if progress:
                        subtotal = model.subtreeCount(doc)
                        if subtotal != 1:
                            total += subtotal - 1
                            ctx.update(total=total)
                    model.remove(doc, progress=ctx)
                    if progress:
                        current += subtotal
                        if ctx.progress._payload['data']['current'] != current:
                            ctx.update(current=current, message='Deleted ' + kind)

    def _getResource(self, id, type):
        model = self._getResourceModel(type)
        return model.load(id=id, user=self.getCurrentUser(), level=AccessType.READ)

    @access.admin
    @autoDescribeRoute(
        Description('Get any resource by ID.')
        .param('id', 'The ID of the resource.', paramType='path')
        .param('type', 'The type of the resource (item, file, etc.).')
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the resource.', 403)
    )
    def getResource(self, id, type):
        return self._getResource(id, type)

    @access.admin
    @autoDescribeRoute(
        Description('Set the created or updated timestamp for a resource.')
        .param('id', 'The ID of the resource.', paramType='path')
        .param('type', 'The type of the resource (item, file, etc.).')
        .param('created', 'The new created timestamp.', required=False)
        .param('updated', 'The new updated timestamp.', required=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Access was denied for the resource.', 403)
    )
    def setTimestamp(self, id, type, created, updated):
        user = self.getCurrentUser()
        model = self._getResourceModel(type)
        doc = model.load(id=id, user=user, level=AccessType.WRITE, exc=True)

        if created is not None:
            if 'created' not in doc:
                raise RestException('Resource has no "created" field.')
            doc['created'] = parseTimestamp(created)
        if updated is not None:
            if 'updated' not in doc:
                raise RestException('Resource has no "updated" field.')
            doc['updated'] = parseTimestamp(updated)
        return model.filter(model.save(doc), user=user)

    def _prepareMoveOrCopy(self, resources, parentType, parentId):
        user = self.getCurrentUser()
        self._validateResourceSet(resources, ('folder', 'item'))

        if resources.get('item') and parentType != 'folder':
            raise RestException('Invalid parentType.')
        return ModelImporter.model(parentType).load(
            parentId, level=AccessType.WRITE, user=user, exc=True)

    @access.user(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Move a set of items and folders.')
        .jsonParam('resources', 'A JSON-encoded set of resources to move. Each type '
                   'is a list of ids.  Only folders and items may be specified.  '
                   'For example: {"item": [(item id 1), (item id2)], "folder": '
                   '[(folder id 1)]}.', requireObject=True)
        .param('parentType', 'Parent type for the new parent of these resources.',
               enum=('user', 'collection', 'folder'))
        .param('parentId', 'Parent ID for the new parent of these resources.')
        .param('progress', 'Whether to record progress on this task.',
               required=False, default=False, dataType='boolean')
        .errorResponse('Unsupported or unknown resource type.')
        .errorResponse('Invalid resources format.')
        .errorResponse('Resource type not supported.')
        .errorResponse('No resources specified.')
        .errorResponse('Resource not found.')
        .errorResponse('ID was invalid.')
    )
    def moveResources(self, resources, parentType, parentId, progress):
        user = self.getCurrentUser()
        parent = self._prepareMoveOrCopy(resources, parentType, parentId)
        total = sum([len(resources[key]) for key in resources])
        with ProgressContext(
                progress, user=user, title='Moving resources',
                message='Calculating requirements...', total=total) as ctx:
            for kind in resources:
                model = self._getResourceModel(kind, 'move')
                for id in resources[kind]:
                    doc = model.load(id=id, user=user, level=AccessType.WRITE, exc=True)
                    ctx.update(message='Moving %s %s' % (kind, doc.get('name', '')))
                    if kind == 'item':
                        if parent['_id'] != doc['folderId']:
                            model.move(doc, parent)
                    elif kind == 'folder':
                        if ((parentType, parent['_id'])
                                != (doc['parentCollection'], doc['parentId'])):
                            model.move(doc, parent, parentType)
                    ctx.update(increment=1)

    @access.user(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Copy a set of items and folders.')
        .jsonParam('resources', 'A JSON-encoded set of resources to copy. Each type '
                   'is a list of ids.  Only folders and items may be specified.  '
                   'For example: {"item": [(item id 1), (item id2)], "folder": '
                   '[(folder id 1)]}.', requireObject=True)
        .param('parentType', 'Parent type for the new parent of these '
               'resources.')
        .param('parentId', 'Parent ID for the new parent of these resources.')
        .param('progress', 'Whether to record progress on this task.',
               required=False, default=False, dataType='boolean')
        .errorResponse('Unsupported or unknown resource type.')
        .errorResponse('Invalid resources format.')
        .errorResponse('Resource type not supported.')
        .errorResponse('No resources specified.')
        .errorResponse('Resource not found.')
        .errorResponse('ID was invalid.')
    )
    def copyResources(self, resources, parentType, parentId, progress):
        user = self.getCurrentUser()
        parent = self._prepareMoveOrCopy(resources, parentType, parentId)
        total = len(resources.get('item', []))
        if 'folder' in resources:
            model = self._getResourceModel('folder')
            for id in resources['folder']:
                folder = model.load(id=id, user=user, level=AccessType.READ, exc=True)
                total += model.subtreeCount(folder)
        with ProgressContext(
                progress, user=user, title='Copying resources',
                message='Calculating requirements...', total=total) as ctx:
            for kind in resources:
                model = self._getResourceModel(kind)
                for id in resources[kind]:
                    doc = model.load(id=id, user=user, level=AccessType.READ, exc=True)
                    ctx.update(message='Copying %s %s' % (kind, doc.get('name', '')))
                    if kind == 'item':
                        model.copyItem(doc, folder=parent, creator=user)
                        ctx.update(increment=1)
                    elif kind == 'folder':
                        model.copyFolder(
                            doc, parent=parent, parentType=parentType, creator=user, progress=ctx)



import cherrypy
import logging
import re
import time


LoggingFilters = []
SingletonRegexLoggingFilter = None


class RegexLoggingFilter(logging.Filter):
    """
    Check log messages against a list of compiled regular expressions.  If any
    of them match, throttle logs.
    """

    def filter(self, record):
        if getattr(record, '_logging_filter', None) is not None:
            return record._logging_filter
        record._logging_filter = True
        msg = record.getMessage()
        for filter in LoggingFilters:
            if filter['re'].search(msg):
                filter['count'] += 1
                if ((filter['frequency']
                        and filter['count'] >= filter['frequency'])
                        or (filter['duration']
                            and time.time() - filter.get('timestamp', 0) > filter['duration'])):
                    if filter['count'] > 1:
                        record.msg += ' (%d similar messages)' % filter['count']
                    filter['count'] = 0
                    filter['timestamp'] = time.time()
                    return True
                record._logging_filter = False
                return False
        return True


def addLoggingFilter(regex, frequency=None, duration=None):
    """
    Add a regular expression to the logging filter.  If the regular expression
    matches a registered regex exactly, just update the frequency value.

    :param regex: a regular expression to match against log messages.  For
        matching cherrypy endpoint logging, this should probably be something
        like 'GET /api/v1/item/[0-9a-fA-F]+/download[/ ?#]'.   More generally,
        a value like GET (/[^/ ?#]+)*/item/[^/ ?#]+/download[/ ?#] would be
        agnostic to the api_root.
    :param frequency: either None to never log matching log messages, or an
        integer, where one log message is emitted out of the specified number.
    :param duration: either None to not log based on elapsed time, or a float
        value of seconds between logging.
    """
    # Always make sure that cherrypy is using our logging filter class.  This
    # is done as a singleton so that addFilter will not duplicate the filter.
    # By doing this here, the import order doesn't matter, and additional
    # cherrypy handlers can be added after import, provided that
    # addLoggingFilter is called after the new logging handlers were added.
    global SingletonRegexLoggingFilter

    if not SingletonRegexLoggingFilter:
        SingletonRegexLoggingFilter = RegexLoggingFilter()
    for handler in cherrypy.log.access_log.handlers:
        handler.addFilter(SingletonRegexLoggingFilter)
    # Also apply to the root handler
    for handler in logging.getLogger().handlers:
        handler.addFilter(SingletonRegexLoggingFilter)

    # Now add or adjust the regex filter.
    newFilter = None
    for filter in LoggingFilters:
        if filter['regex'] == regex:
            newFilter = filter
    if not newFilter:
        newFilter = {
            'regex': regex,
            're': re.compile(regex),
            'count': 0
        }
        LoggingFilters.append(newFilter)
    newFilter['frequency'] = frequency
    newFilter['duration'] = duration


def removeLoggingFilter(regex):
    """
    Remove a regular expression from the logging filter.

    :param regex: the regular expression to remove.
    :returns: True if a filter was removed.
    """
    for idx in range(len(LoggingFilters)):
        if LoggingFilters[idx]['regex'] == regex:
            LoggingFilters[idx:idx + 1] = []
            return True
    return False



import asyncio
import datetime
import functools
import json
import logging
import os
import time
import uuid

import redis
import redis.asyncio as aioredis
from starlette.endpoints import WebSocketEndpoint
from starlette.websockets import WebSocket

from girder.constants import TokenScope
from girder.models.token import Token

logger = logging.getLogger(__name__)


class ProgressState:
    """
    Enum of possible progress states for progress records.
    """

    QUEUED = 'queued'
    ACTIVE = 'active'
    SUCCESS = 'success'
    ERROR = 'error'

    @classmethod
    def isComplete(cls, state):
        return state == cls.SUCCESS or state == cls.ERROR


@functools.lru_cache
def _redis_client_async() -> aioredis.Redis:
    url = os.environ.get('GIRDER_NOTIFICATION_REDIS_URL', 'redis://localhost:6379')
    return aioredis.Redis.from_url(url)


@functools.lru_cache
def _redis_client_sync() -> redis.Redis:
    url = os.environ.get('GIRDER_NOTIFICATION_REDIS_URL', 'redis://localhost:6379')
    return redis.Redis.from_url(url)


class UserNotificationsSocket(WebSocketEndpoint):
    user_id: str

    async def on_connect(self, websocket):
        token_id = websocket.query_params.get('token')
        if not token_id:
            await websocket.close(code=3000, reason='Token is required')
            return

        token = Token().load(token_id, force=True, objectId=False)
        if (
            token is None
            or token['expires'] < datetime.datetime.now(datetime.timezone.utc)
            or 'userId' not in token
            or not Token().hasScope(token, TokenScope.USER_AUTH)
        ):
            await websocket.close(code=3000, reason='Invalid token')
            return

        await websocket.accept()

        self.user_id = token['userId']
        self.pubsub = _redis_client_async().pubsub()
        await self.pubsub.subscribe(f'user_{self.user_id}')

        self.listen_task = asyncio.create_task(self.listen_and_forward(websocket))

    async def listen_and_forward(self, websocket: WebSocket):
        try:
            async for message in self.pubsub.listen():
                if message['type'] == 'message':
                    await websocket.send_text(message['data'].decode())
        except asyncio.CancelledError:
            pass
        finally:
            await self.pubsub.unsubscribe()
            await self.pubsub.close()

    async def on_disconnect(self, websocket, close_code):
        if hasattr(self, 'listen_task'):
            self.listen_task.cancel()
            await self.listen_task


class Notification:
    def __init__(self, type: str, data: dict, user: dict, **payload):
        """
        Create a notification for a specific user's notification channel.

        :param type: The notification type.
        :param data: The notification payload.
        :param user: User to send the notification to.
        """
        self._payload = {
            'type': type,
            'data': data,
            **payload,
        }
        self._user = user

    def flush(self):
        msg = json.dumps(self._payload, default=str)

        try:
            _redis_client_sync().publish(f'user_{self._user["_id"]}', msg)
        except redis.RedisError:
            logger.exception('Error flushing notification to redis')

    @classmethod
    def initProgress(cls, user, title, total=0, state=ProgressState.ACTIVE,
                     current=0, message='', estimateTime=True, resource=None,
                     resourceName=None, _id=None) -> 'Notification':
        """
        Create a "progress" type notification that can be updated anytime there
        is progress on some task. It is the caller's responsibility to call `flush`
        on the returned object to actually send the notification.

        :param user: the user associated with this notification.  If this is
            None, a session token must be specified.
        :param title: The title of the task. This should not change over the
            course of the task. (e.g. 'Deleting folder "foo"')
        :type title: str
        :param total: Some numeric value representing the total task length. By
            convention, setting this <= 0 means progress on this task is
            indeterminate.
        :type total: int, long, or float
        :param state: Represents the state of the underlying task execution.
        :type state: ProgressState enum value.
        :param current: Some numeric value representing the current progress of
            the task (relative to total).
        :type current: int, long, or float
        :param message: Message corresponding to the current state of the task.
        :type message: str
        :param token: if the user is None, associate this notification with the
            specified session token.
        :param estimateTime: if True, generate an estimate of the total time
            the task will take, if possible.  If False, never generate a time
            estimate.
        :param resource: a partial or complete resource that the notification is
            associated with. This must at a minimum include the id of the resource.
        :param resourceName: the type of resource the notification is associated with.
        :param _id: the unique ID of this progress stream. If not provided, a random UUID will
            be used. Only pass this if you need to update the same progress stream from multiple
            request contexts or task contexts, where it's not possible to reuse the same
            instance of this class.
        """
        if _id is None:
            _id = str(uuid.uuid4())

        data = {
            'title': title,
            'total': total,
            'current': current,
            'state': state,
            'message': message,
            'resource': resource,
            'resourceName': resourceName,
        }

        return cls(
            'progress', data, user, estimateTime=estimateTime, startTime=time.time(), _id=_id
        )

    def updateProgress(self, *, increment: int = None, **kwargs):
        """
        Send a progress update message to any listeners of this notification.

        :param increment: The amount to increment the current progress by.
        :param kwargs: Any other fields to update in the progress message payload.
        """
        if increment is not None:
            self._payload['data']['current'] += increment

        self._payload['data'].update(kwargs)

        current, total = self._payload['data']['current'], self._payload['data']['total']

        if self._payload['estimateTime'] and total > 0 and current > 0:
            self._payload['updatedTime'] = time.time()
            self._payload['estimatedTotalTime'] = (
                total * (self._payload['updatedTime'] - self._payload['startTime']) / current
            )

        self.flush()



import importlib.metadata
import logging
import os

from .utility import config

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    # package is not installed
    __version__ = None

__license__ = 'Apache 2.0'
auditLogger = logging.getLogger('girder_audit')
auditLogger.setLevel(logging.INFO)
auditLogger.propagate = False

if not os.getenv('GIRDER_SPHINX_BUILD'):
    config.loadConfig()  # Populate the config info at import time
